
/*
** Copyright (c) 2006 Herbert Poetzl
**
** V0.1 peer cred test
** V0.2 peer cred set/verify
** V0.3 unix sockets
**
*/

#define VERSION "V0.3"

#include <stdio.h>      /* for printf() and fprintf() */
#include <sys/socket.h> /* for socket(), bind(), and connect() */
#include <sys/un.h>	/* for sockaddr_un */
#include <arpa/inet.h>  /* for sockaddr_in and inet_ntoa() */
#include <stdlib.h>     /* for atoi() and exit() */
#include <string.h>     /* for memset() */
#include <fcntl.h>      /* for fcntl() */
#include <unistd.h>     /* for close() */
#include <netinet/in.h>	/* for inet_aton */
#include <netdb.h>
#include <errno.h>


static	char *	host = "localhost";
static	char *	path = "/tmp/socket";
static	int	port = 80;
static	int	timeout = 10;

static	struct	in_addr addr = { 0 };
static	int	euid = 0;
static	int	egid = 0;

static	char	err_flag = 0;

static	char	opt_server = 0;
static	char	opt_unix = 0;
static	char	opt_dgram = 0;
static	char	opt_setuid = 0;
static	char	opt_setgid = 0;

static	char *	cmd_name;


#define SO_PEERTAG 32

struct peer_tag {
        int32_t xid;
        int32_t nid;
};


int	check_creds(char *label, int sock, int pf)
{
	union {
	    struct sockaddr_in ip;
	    struct sockaddr_un up;
	    } peer;
	socklen_t peerlen = sizeof(peer);
	int ret;

	struct ucred peercred = { 0,0,0 };
	socklen_t peercredlen = sizeof(peercred);
	struct peer_tag peertag = { 0,0 };
	socklen_t peertaglen = sizeof(peertag);

	memset(&peer, 0, peerlen);	/* Zero out structure */
	ret = getpeername(sock, (struct sockaddr *)&peer, &peerlen);
	if (ret)
	    perror("getpeername");

	ret = getsockopt(sock, SOL_SOCKET,
	    SO_PEERCRED, &peercred, &peercredlen);
	if (ret)
	    perror("getsockopt [SO_PEERCRED]");

	ret = getsockopt(sock, SOL_SOCKET,
	    SO_PEERTAG, &peertag, &peertaglen);
	if (ret)
	    perror("getsockopt [SO_PEERTAG]");
	
	if (pf == PF_INET) {
	    printf("%12s: %s:%u tag=%u:%u\n",
	    	label, inet_ntoa(peer.ip.sin_addr), 
		peer.ip.sin_port, peertag.xid, peertag.nid);
	} else if (pf == PF_UNIX) {
	    printf("%12s: %s pid=%u %d:%d tag=%u:%u\n",
	    	label, peer.up.sun_path,
	    	peercred.pid, peercred.uid, peercred.gid,
	    	peertag.xid, peertag.nid);
	} else {
	    printf("%12s: tag=%u:%u\n",
	    	label, peertag.xid, peertag.nid);
	}
	return 0;
}



int	accept_inet(int sock)
{
	struct sockaddr_in from;
	socklen_t len = sizeof(from);
	int csock;

	/* Accept socket connection */
	csock = accept(sock, (struct sockaddr *)&from, &len);
	if (csock < 0) {
	    perror("accept");
	}

	check_creds("client-inet", csock, PF_INET);

	/* Close socket connection */
	close(csock);
	return 0;
}

int	accept_unix(int sock)
{
	struct sockaddr_un from;
	socklen_t len = sizeof(from);
	int csock;

	/* Accept socket connection */
	csock = accept(sock, (struct sockaddr *)&from, &len);
	if (csock < 0) {
	    perror("accept");
	}

	check_creds("client-unix", csock, PF_UNIX);

	/* Close socket connection */
	close(csock);
	return 0;
}


#define MAX_PENDING	5	/* Maximum outstanding connection requests */

int	server_inet(int type, int proto)
{
	struct sockaddr_in saddr;
	int sock;

	/* Create socket for incoming connections */
	if ((sock = socket(PF_INET, type, proto)) < 0) {
	    perror("socket");
	    exit(1);
	}

	/* Construct address structure */
	memset(&saddr, 0, sizeof(saddr));	/* Zero out structure */
	saddr.sin_family = AF_INET;		/* Internet address family */
	saddr.sin_addr = addr;			/* Server address */
	saddr.sin_port = htons(port);		/* Local port */

	/* Bind to the local address */
	if (bind(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
	    perror("bind");
	    exit(2);
	}

	/* Mark the socket so it will listen for incoming connections */
	if (listen(sock, MAX_PENDING) < 0) {
	    perror("listen");
	    exit(3);
	}
	
	while (1) {
	    fd_set read_set;	/* Set of socket descriptors for select() */
	    struct timeval sel_tout; 

	    FD_ZERO(&read_set);
	    FD_SET(STDIN_FILENO, &read_set);
	    FD_SET(sock, &read_set);

	    sel_tout.tv_sec = timeout;
	    sel_tout.tv_usec = 0;

	    if (select(sock + 1, &read_set, NULL, NULL, &sel_tout) == 0) {
		printf(" still alive.\n");
	    } else {
		if (FD_ISSET(0, &read_set)) {
		    /* terminal input */
		    printf(" keyboard input.\n");
		    break;
		}
		if (FD_ISSET(sock, &read_set)) {
		    /* socket connection */
		    accept_inet(sock);
		}
	    }
	}

	/* Close socket connection */
	close(sock);
	return 0;
}


int	server_unix(int type)
{
	struct sockaddr_un saddr;
	int len, sock;

	/* Create socket for incoming connections */
	if ((sock = socket(PF_UNIX, type, 0)) < 0) {
	    perror("socket");
	    exit(1);
	}

	/* Construct address structure */
	memset(&saddr, 0, sizeof(saddr));	/* Zero out structure */
	saddr.sun_family = AF_UNIX;		/* Unix address family */
	strncpy(saddr.sun_path, path, sizeof(saddr.sun_path));
	len = strlen(saddr.sun_path) + sizeof(saddr.sun_family);
	
	/* Remove existing socket */
	unlink(saddr.sun_path);

	/* Bind to the path */
	if (bind(sock, (struct sockaddr *)&saddr, len) < 0) {
	    perror("bind");
	    exit(2);
	}

	/* Mark the socket so it will listen for incoming connections */
	if (listen(sock, MAX_PENDING) < 0) {
	    perror("listen");
	    exit(3);
	}
	

	while (1) {
	    fd_set read_set;	/* Set of socket descriptors for select() */
	    struct timeval sel_tout; 

	    FD_ZERO(&read_set);
	    FD_SET(STDIN_FILENO, &read_set);
	    FD_SET(sock, &read_set);

	    sel_tout.tv_sec = timeout;
	    sel_tout.tv_usec = 0;

	    if (select(sock + 1, &read_set, NULL, NULL, &sel_tout) == 0) {
		printf(" still alive.\n");
	    } else {
		if (FD_ISSET(0, &read_set)) {
		    /* terminal input */
		    printf(" keyboard input.\n");
		    break;
		}
		if (FD_ISSET(sock, &read_set)) {
		    /* socket connection */
		    accept_unix(sock);
		}
	    }
	}

	/* Close socket connection */
	close(sock);
	/* Remove created socket */
	unlink(saddr.sun_path);
	return 0;
}


int	client_inet(int type, int proto)
{
	struct sockaddr_in peer;
	int sock;

	if ((sock = socket(PF_INET, type, proto)) < 0) {
	    perror("socket");
	    exit(-1);
	}

	/* Construct the peer address structure */
	memset(&peer, 0, sizeof(peer));		/* Zero out structure */
	peer.sin_family = AF_INET;		/* Internet address family */
	peer.sin_addr = addr;			/* Peer IP address */
	peer.sin_port = htons(port);		/* Peer port */

	/* Establish the connection to the peer */
        if (connect(sock, (struct sockaddr *)&peer, sizeof(peer)) < 0) {
	    perror("connect");
	    exit(-2);
	}

	check_creds("peer-inet", sock, PF_INET);

	/* Close socket connection */
	close(sock);
	return 0;
}

int	client_unix(int type)
{
	struct sockaddr_un peer;
	int len, sock;

	if ((sock = socket(PF_UNIX, type, 0)) < 0) {
	    perror("socket");
	    exit(-1);
	}

	/* Construct address structure */
	memset(&peer, 0, sizeof(peer));		/* Zero out structure */
	peer.sun_family = AF_UNIX;		/* Unix address family */
	strncpy(peer.sun_path, path, sizeof(peer.sun_path));
	len = strlen(peer.sun_path) + sizeof(peer.sun_family);

	/* Establish the connection to the peer */
        if (connect(sock, (struct sockaddr *)&peer, len) < 0) {
	    perror("connect");
	    exit(-2);
	}

	check_creds("peer-unix", sock, PF_UNIX);

	/* Close socket connection */
	close(sock);
	return 0;
}




#define	OPTIONS	"hdsu:g:I:P:T:U:"

int	main(int argc, char *argv[])
{
	extern int optind;
	extern char *optarg;
	char c, errflg = 0;

	cmd_name = argv[0];
	while ((c = getopt(argc, argv, OPTIONS)) != EOF) {
	    switch (c) {
	    case 'h':
		fprintf(stderr,
		    "This is %s " VERSION "\n"
		    "options are:\n"
		    "-h        print this help message\n"
		    "-s        act as server\n"
		    "-d        use datagrams\n"
		    "-u <uid>  effective uid\n"
		    "-g <gid>  effective gid\n"
		    "-I <ip>   ip to use/query [%s]\n"
		    "-P <port> port to use/query [%d]\n"
		    "-T <secs> server timeout [%d]\n"
		    "-U <path> use unix socket\n"
		    ,cmd_name, host, port, timeout);
		exit(0);
		break;
	    case 's':
		opt_server = 1;
		break;
	    case 'd':
		opt_dgram = 1;
		break;
	    case 'u':
	    	opt_setuid = 1;
		euid = atoi(optarg);
		break;
	    case 'g':
	    	opt_setgid = 1;
		egid = atoi(optarg);
		break;
	    case 'I':
		host = optarg;
		break;
	    case 'P':
		port = atol(optarg);
		break;
	    case 'T':
	    	timeout = atoi(optarg);
		break;
	    case 'U':
	    	opt_unix = 1;
		path = optarg;
		break;
	    case '?':
	    default:
		errflg++;
		break;
	    }
	}
	if (errflg) {
	    fprintf(stderr,
		"Usage: %s -[" OPTIONS "]\n"
		"%s -h for help.\n",
		cmd_name, cmd_name);
	    exit(2);
	}

	inet_aton(host, &addr);
	
	if (opt_setgid) {
	    if (setegid(egid))
	    	perror("setegid");
	}
	if (opt_setuid) {
	    if (seteuid(euid))
	    	perror("seteuid");
	}
	
	if (opt_unix) {
	    int type = opt_dgram ? SOCK_DGRAM : SOCK_STREAM;

	    if (opt_server)
	    	server_unix(type);
	    else
	    	client_unix(type);
	} else {
	    int type = opt_dgram ? SOCK_DGRAM : SOCK_STREAM;
	    int proto = opt_dgram ? IPPROTO_UDP : IPPROTO_TCP;
	    
	    if (opt_server)
	    	server_inet(type, proto);
	    else
	    	client_inet(type, proto);
	}

	exit((err_flag)?1:0);
}

